##
## @file fnlQLearn.R
## @brief Contains fns for estimating optimal treatment regimes
##        via Q-learning with a single fnl predictor
## @author Anonymized
## @note Anonymized
##

##
## @fn denseQLearn (X, W, A, Y)
## @brief Implements fnl Q-learning at a single stage.  Note that
##        this functional works for both dense and sparse data, the
##        name is a holdover from preliminary simulations with
##        dense data. 
## @param X n x p matrix of predictors (includes intercept)
## @param W n x nObs matrix of observed fnl values
## @param A n vector of txts in {-1,1}
## @param Y cts outcome, coded s.t. higher is better
## @return a list with the following components:
##         alphaHat-->est coeff vector for main effect of X
##         nuHat-->est coeff vector linear probe of W
##         gammaHat-->est coeff vector for interaction of X and A
##         etaHat-->est coeff vector for interaction of lin probe of W and A
##         piHat-->function from X, W to a recommended txt 
##
denseQLearn = function (X, W, A, Y){
    require ("refund");
    fpcaFit = fpca.sc (Y=W,var=TRUE, pve=0.99);
    lenT = dim(W)[2]; 
    ldings = fpcaFit$scores/sqrt(lenT);
    dMat = cbind(X, ldings, A*X, A*ldings);
    thetaHat = lm.fit (dMat, Y)$coefficients;
    p = dim (X)[2];
    alphaHat = thetaHat[1:p];
    d = dim(ldings)[2];
    nuHat = thetaHat[(p+1):(p+d)];
    gammaHat = thetaHat[(p+d+1):(2*p+d)];
    etaHat = thetaHat[(2*p+d+1):(2*p+2*d)];

    getLoadings = function (WNew){
        ldv = t(t(WNew)-as.numeric(fpcaFit$mu)) %*% fpcaFit$efunctions/
            sqrt(lenT);
        return (ldv);
    }

    Phi = fpcaFit$efunctions;
    estCoefFn = Phi %*% etaHat*sqrt(lenT);
    Lam = fpcaFit$evalues;
    vig = fpcaFit$sigma2;
    getLoadingsSprs = function (WNew){
        wn = dim(WNew)[1];
        nLdings = matrix (0, nrow=wn, ncol=d);
        for (i in 1:wn){
            ind = which(!is.na(WNew[i,]));
            m = length(ind);
            Phii = Phi[ind,];
            Im = diag (m);
            dgs = vig^(-1)*Im - vig^(-1)*Phii %*%
                solve (vig*diag(1/Lam) + t(Phii) %*% Phii) %*% t(Phii);
            nLdings[i,] = Lam*t(Phii) %*% dgs %*%
                (WNew[i,ind] - as.numeric(fpcaFit$mu[ind]))/sqrt(lenT);
        }
        return (nLdings);
    }
    
    getTxts = function (XNew, WNew){
        nLdings = getLoadingsSprs (WNew);
        pDMat = cbind (XNew, nLdings, XNew, nLdings);
        nDMat = cbind (XNew, nLdings, -XNew, -nLdings);
        return (sign((pDMat-nDMat)%*%thetaHat));
    }
    return (list("alphaHat"= alphaHat,
                 "nuHat"=nuHat,
                 "gammaHat"=gammaHat,
                 "etaHat"=etaHat,
                 "piHat"=getTxts,
                 "getLdings"=getLoadingsSprs,
                 "coefFn"=estCoefFn,
                 "Phi"=Phi,
                 "mu"=fpcaFit$mu));
}


## Same as above but does not make use of fnl predictors 
stdQLearn = function (X, W, A, Y){
    nObs = dim(W)[2];
    n = dim(W)[1];
    ## Last obs
    maxInd = cbind (1:n, unlist (lapply (apply (!is.na(W), 1, which), max)));
    dMatLast = cbind (X, W[maxInd], A*X, A*W[maxInd]);
    thetaLast = lm.fit (dMatLast, Y)$coefficients;
    getTxtsLast = function (XNew, WNew){
        nn = dim(WNew)[1];
        nMaxInd = cbind(1:nn,
            unlist (lapply (apply(!is.na(WNew), 1, which), max)));
        pDMat = cbind (XNew, WNew[nMaxInd], XNew, WNew[nMaxInd]);
        nDMat = cbind (XNew, WNew[nMaxInd], -XNew, -WNew[nMaxInd]);
        return (sign ((pDMat-nDMat)%*%thetaLast));
    }

    ## Mean obs
    WBar = apply(W, 1, mean, na.rm=T);
    dMatMean = cbind (X, WBar, A*X, A*WBar);
    thetaMean = lm.fit (dMatMean, Y)$coefficients;
    getTxtsMean = function (XNew, WNew){
        WNewMean = apply (WNew, 1, mean, na.rm=T);
        pDMat = cbind (XNew, WNewMean, XNew, WNewMean);
        nDMat = cbind (XNew, WNewMean, -XNew, -WNewMean);
        return (sign((pDMat - nDMat)%*%thetaMean));
    }
    return (list('piHatLast'=getTxtsLast,
                'piHatMean'=getTxtsMean));
}
