##
## @fn genSparseDataFixed.R
## @brief Generates sparsely observed longitudinal data
##        for use in txt selection.  The model is of the form:
##        Y = x^{\T}\alpha + f(Z) +
##              a(x^{\T}\gamma + f(Z)) + \epsilon
##        where f(Z) is known functional of Z
##        epsilon is a normal error term.
## @author Anonymized
## @note Anonymized
##

##
## @fn genSparseFixed (n, obsPts, smplPts, nBasis, fz, bLoadings,
##                    p, alpha, gamma, nu, sigma, wnSigma, lam, kappa)
## @param n number of points to generate
## @param obsPts equaly spaced points on which *all*
##        observation times will fall, these must be in
##        the range (0,1)
## @param smplPts equally spaced points on which to
##        evaluate Z for evaluation of f(Z) 
## @param nBasis the number of basis functions 
## @param fz functional defined on L^2(0,1), takes
##        two arguments: z, tz corresponding to z evaluated
##        at the points tz, returns a scalar
## @param p dimension of static predictors x
## @param alpha main effect coeffs for static predictors
## @param gamma interaction coeffs for static predictors
## @param nu scalar coeff for fz in main effect
## @param eta scalar coeff for fz in interaction
## @param sigma residual variance
## @param wnSigma white noise process variance
## @param kappaMin minimum number of pts to sample for each subj
## @param kappaMax maximum number of pts to sample for each subj
## @return list with the following key-->value pairs:
##         X --> matrix static predictors
##         W --> matrix of functional predictors 
##         A --> vector of txts
##         Y --> vector or responses
##
genSparseDataFixed = function (n, obsPts, smplPts, nBasis, fz, p, alpha,
    gamma, nu, eta, sigma, wnSigma, kappaMin, kappaMax){
    require ('fda');
    ## Generate non-fnl predictors
    X = cbind(1, matrix (rnorm(n*(p-1)), nrow=n));
    A = sign (rnorm(n));
    
    ## Generate fnl predictors
    if (max(obsPts) > 1 | min(obsPts) < 0){
        stop ('obsPts must belong to [0,1]');
    }
    bsb = create.bspline.basis (range=c(0,1), nbasis=nBasis);
    W = matrix (0, nrow=n, ncol=length(obsPts));
    lFz = numeric (n);
    ZAll = matrix (0, nrow=n, ncol=length(smplPts));
    
    mu = rep(0, length(obsPts));
    for (i in 1:n){
        randCoeff = rnorm (nBasis);
        z = mu + eval.basis (obsPts, bsb) %*% randCoeff;
        zFull = mu + eval.basis (smplPts, bsb) %*% randCoeff;
        ZAll[i,] = zFull;
        lFz[i] = fz (zFull, smplPts); 
        err = rnorm(length(obsPts))*wnSigma;
        W[i,] = z + err;
        if (kappaMin < kappaMax){
            nSamples = length(obsPts) - sample (kappaMin:kappaMax, size=1);
        }
        else{
            nSamples = length(obsPts) - kappaMax;
        }
        W[i,sample(1:length(obsPts), size=nSamples)] = NA;  
    }
    Y = X %*% alpha + nu*lFz + A*(X%*%gamma + eta*lFz) + rnorm(n)*sigma;
    optA = sign (X%*%gamma + eta*lFz);
    return (list(X=X, W=W, A=A, Y=Y, fzVals=lFz, fz=fz, optA=optA,
                 Z=ZAll));
}


usage = function (){
    set.seed (1);
    n = 100;
    nObsPts = 50;
    nSmplPts = 1000;
    obsPts = seq (0, 1, length=nObsPts);
    smplPts = seq (0, 1, length=nSmplPts);
    nBasis = 4;
    ## (Approximate) total variation distance
    fz = function (z, tz){
        zsort = sort (z);
        return (sum(abs(diff(z))));
    }
    
    p = 5;
    alpha = rep (1, p);
    gamma = rep (1, p);
    nu = 1;
    eta = -0.40;
    sigma=1.0;
    wnSigma = 0.15;
    kappaMin = 5;
    kappaMax = 15;
    op = genSparseDataFixed (n, obsPts, smplPts, nBasis, fz, p, alpha,
        gamma, nu, eta, sigma, wnSigma, kappaMin, kappaMax);
    source ("fnlQLearnFixed.R");
}





